#pragma once

#include "v3.h"
#include "framebuffer.h"
#include "ppc.h"

class TMesh {
public:
  bool disabled;
  V3 *verts, *cols, *normals, *tcs; // vertex coordinates, vertex colors, vertex normals
  int vertsN; // number of vertices
  V3 *projVerts;
  FrameBuffer *texture; // texture is just a pointer; TMesh does not own texture memory
  unsigned int *tris; // connectivity data; triangle i has 
      // vertices verts[tris[3*i+0]], verts[tris[3*i+1]], verts[tris[3*i+2]]
  int trisN; // number of triangles (not number of indices in array)
  TMesh() : texture(0), disabled(false), tcs(0), projVerts(0), verts(0), cols(0), normals(0), vertsN(0), tris(0), trisN(0) {};
  void LoadBin(char *fname); // load from bin file
  void SetQuad(V3 *_verts, V3 *_cols);
  void Allocate(int _vertsN, int _trisN);

  // drawing
  void RenderWireframe(PPC *ppc, FrameBuffer *fb);
  void RenderFilled(PPC *ppc, FrameBuffer *fb);
  void RenderHW();
  void RenderRT(PPC *ppc, FrameBuffer *fb);
  bool RenderRT(V3 rO, V3 rDir, unsigned int &color, float &z, V3 &ip, V3 &rr);

  // transformations
  void Translate(V3 tv);

  // average of all vertices;
  V3 GetCenter();
  AABB ComputeAABB();

  // ray tracing
  bool IntersectTriangleWithRay(int trisi, V3 rO, V3 rDir, V3& abt);


};