#include "kernel.h"

#include "framebuffer.h"

void Kernel::SetAverage(int _n) {

  n = _n;
  krn = new float[n*n];
  for (int i = 0; i < n*n; i++)
    krn[i] = 1.0f;
}

void Kernel::Normalize() {

  float sum = 0.0f;
  for (int i = 0; i < n*n; i++)
    sum += krn[i];

  for (int i = 0; i < n*n; i++)
    krn[i] /= sum;

}

float Kernel::Sum() {

  float sum = 0.0f;
  for (int i = 0; i < n*n; i++) {
    sum += krn[i];
  }

  return sum;
  
}


void Kernel::SetEdgeDetection() {

  n = 3;
  krn = new float[n*n];
  krn[0] = 0.0f;
  krn[2] = 0.0f;
  krn[6] = 0.0f;
  krn[8] = 0.0f;

  krn[1] = -1.0f;
  krn[3] = -1.0f;
  krn[5] = -1.0f;
  krn[7] = -1.0f;

  krn[4] = 4.0f;

}

void Kernel::Convolve(FrameBuffer *infb, FrameBuffer *outfb) {

  if (outfb->w != infb->w || outfb->h != infb->h) {
    cerr << "ERROR: output image of wrong size in Kernel::Convolve" << endl;
    return;
  }

  outfb->Clear(0xFF000000, 0.0f);
  float sum = Sum();
  for (int v = n/2; v < outfb->h-n/2; v++) {
    for (int u = n/2; u < outfb->w-n/2; u++) {
      V3 outv(0.0f, 0.0f, 0.0f);
      for (int i = 0; i < n; i++) {
        for (int j = 0; j < n; j++) {
          V3 inv;
          inv.SetFromUIntColor(infb->Get(u-n/2+j, v-n/2+i));
          outv = outv + inv*krn[i*n+j];
        }
      }
      outv = outv.Absolute();
      outv = outv.Clamp(0.0f, 1.0f);
      outfb->Set(u, v, outv.UIntColor());
    }
  }

}
