#include "scene.h"
#include "v3.h"
#include "m33.h"
#include <float.h>
#include "kernel.h"
#include <iostream>
#include <fstream>

using namespace std;

Scene *scene;

Scene::Scene() {

  cgi = 0;
  soi = 0;
  animf = 0.0f;

  // create user interface
  gui = new GUI();
  gui->show();

  // create SW framebuffer
  int u0 = 20;
  int v0 = 50;
  int sci = 1;
  int w = sci*480;
  int h = sci*360;
  fb = new FrameBuffer(u0, v0, w, h);
  fb->label("SW Framebuffer");
  rtfb = new FrameBuffer(u0+w+u0, v0, w, h);
  rtfb->label("Ray Tracing Framebuffer");
  rtfb->isRT = true;
  hwfb = new FrameBuffer(u0+w+u0, v0, w, h);
  hwfb->isHW = true;
  hwfb->label("Fixed Pipeline Framebuffer");
  gpufb = new FrameBuffer(u0+w+u0+w+u0, v0, w, h);
  gpufb->isGPU = true;
  gpufb->label("Programmable Pipeline Framebuffer");
  fb3 = new FrameBuffer(u0+w+u0, v0, w, h);
  fb3->label("3rd person");

  lfb = new FrameBuffer(u0+w+u0, v0, w, h);
  lfb->label("Left image");
  rfb = new FrameBuffer(u0+w+u0+u0+w, v0, w, h);
  rfb->label("Right image");
  lppc = rppc = 0;
  cfb = new FrameBuffer(u0, v0+h+v0, w*4/3, h*5/4);
  cfb->label("Canvas");

  tmeshesN = 5;
  tmeshes = new TMesh[tmeshesN];
  tmeshes[0].LoadBin("geometry/teapot1K.bin");
  tmeshes[0].Translate(V3(0.0f, -20.0f, -150.0f));

  V3 verts[4], cols[4];
  cols[0] = cols[1] = V3(1.0f, 1.0f, 0.0f);
  cols[2] = cols[3] = V3(0.0f, 0.0f, 0.0f);

  verts[0] = V3(-50.0f, +40.0f, -120.0f);
  verts[1] = V3(-50.0f, 3.0f, -120.0f);
  verts[2] = V3(+150.0f, 3.0f, -750.0f);
  verts[3] = V3(+150.0f, +40.0f, -750.0f);
  tmeshes[1].SetQuad(verts, cols);

  verts[0] = V3(-50.0f, -3.0f, -120.0f);
  verts[1] = V3(-50.0f, -40.0f, -120.0f);
  verts[2] = V3(+150.0f, -40.0f, -750.0f);
  verts[3] = V3(+150.0f, -3.0f, -750.0f);
  tmeshes[2].SetQuad(verts, cols);

  int texw = 64;
  texture = new FrameBuffer(30, 30, texw, texw);
  unsigned int baw[2] = {0xFF000000, 0xFFAAAAAA};
  texture->SetGrid(baw, 8);
  texture->label("Texture");
//  texture->show();

  verts[0] = V3(-5000.0f, +5000.0f, -100.0f);
  verts[1] = V3(-5000.0f, -5000.0f, -100.0f);
  verts[2] = V3(+5000.0f, -5000.0f, -100.0f);
  verts[3] = V3(+5000.0f, +5000.0f, -100.0f);
  tmeshes[3].SetQuad(verts, cols);
  tmeshes[3].texture = texture;

  V3 tC = tmeshes[0].GetCenter();
  AABB aabb = tmeshes[0].ComputeAABB();
  float th = aabb.lcc[1] - aabb.scc[1];
  float a = 35.0f;
  verts[0] = tC + V3(-a, -th/2.0f, +a);
  verts[1] = tC + V3(+a, -th/2.0f, +a);
  verts[2] = tC + V3(+a, -th/2.0f, -a);
  verts[3] = tC + V3(-a, -th/2.0f, -a);
  cols[0] = V3(1.0f, 0.0f, 0.0f);
  cols[1] = V3(0.0f, 1.0f, 0.0f);
  cols[2] = V3(0.0f, 0.0f, 1.0f);
  cols[3] = V3(1.0f, 1.0f, 0.0f);
  tmeshes[4].SetQuad(verts, cols);
  tmeshes[4].texture = texture;

  tmeshes[0].disabled = false;
  tmeshes[1].disabled = true;
  tmeshes[2].disabled = true;
  tmeshes[3].disabled = true;
  tmeshes[4].disabled = false;

  // create camera
  float hfov = 55.0f;
  ppc = new PPC(hfov, w, h);
  if (!tmeshes[0].disabled)
    ppc->C = tmeshes[0].GetCenter() + V3(0.0f, 0.0f, 110.0f);
  ppc0 = new PPC(hfov, w, h);
  *ppc0 = *ppc;
  ppc1 = new PPC(hfov, w, h);
  ppc1->PositionAndOrient(tmeshes[0].GetCenter() + V3(200.0f, 100.0f, 50.0f),
    tmeshes[0].GetCenter(), V3(0.0f, 1.0f, 0.0f));
  ppc3 = new PPC(90.0f, w, h);
  ppc3->PositionAndOrient(tmeshes[0].GetCenter() + V3(30.0f, 200.0f, 20.0f), 
    tmeshes[0].GetCenter(), V3(0.0f, 1.0f, 0.0f));
  cppc = new PPC(hfov*5/3, cfb->w, cfb->h);
  
  // position UI window
  gui->uiw->position(u0+w*4/3+u0, h+v0+v0);

  fb->show();
//  rtfb->show();
//  hwfb->show();
//  gpufb->show();
//  lfb->show();
//  rfb->show();
//  cfb->show();
//  fb3->show();
  Render();
  RenderHW();
}

void Scene::Render() {

  RenderFirstPerson();
  RenderThirdPerson();


}

void Scene::Render(FrameBuffer *usefb, PPC *useppc) {

  usefb->Clear(0xFFFFFFFF, 0.0f);
  for (int tmi = 0; tmi < tmeshesN; tmi++) {
    tmeshes[tmi].RenderFilled(useppc, usefb);
  }
  usefb->redraw();

}

void Scene::RenderRT(FrameBuffer *usefb, PPC *useppc) {

  usefb->Clear(0xFFFFFFFF, FLT_MAX);
  usefb->redraw();
  Fl::check();
  for (int v = 0; v < usefb->h; v++) {
    for (int i = 0; i < usefb->w; i++)
      usefb->Set(i, v, 0xFFFFFFFF);
    for (int u = 0; u < usefb->w; u++) {
      V3 rO = useppc->C;
      V3 rDir = useppc->c + useppc->a*(.5f + (float)u) + useppc->b*(.5f+(float)v);
      unsigned int color;
      float z;
      V3 ip, rr;
      if (!RayTrace(rO, rDir, color, z, ip, rr))
        continue;
      if (usefb->GetZB(u, v) < z)
        continue;
      usefb->SetZB(u, v, z);
      float rz;
      unsigned int rc;
      V3 rip, rrr;
      if (RayTrace(ip, rr, rc, rz, rip, rrr)) {
        V3 cv;
        cv.SetFromUIntColor(color);
        V3 rcv;
        rcv.SetFromUIntColor(rc);
        color = (rcv*0.3f + cv*0.7f).UIntColor();
      }
      usefb->Set(u, v, color);
    }
    if (v < usefb->h-1) {
      for (int i = 0; i < usefb->w; i++)
        usefb->Set(i, v+1, 0xFF0000FF);
    }
    usefb->redraw();
    Fl::check();
  }

}

bool Scene::RayTrace(V3 rO, V3 rDir, unsigned int &color, 
  float &z, V3 &ip, V3 &rr) {

  bool ret = false;
  z = FLT_MAX;
  unsigned int currcolor;
  float currz;
  V3 currrr, currip;
  for (int tmi = 0; tmi < tmeshesN; tmi++) {
    if (!tmeshes[tmi].RenderRT(rO, rDir, currcolor, currz, currip, currrr))
      continue;
    if (currz < z) {
      z = currz;
      color = currcolor;
      ip = currip;
      rr = currrr;
      ret = true;
    }
  }

  return ret;

}


void Scene::RenderFirstPerson() {

  Render(fb, ppc);

}

void Scene::RenderThirdPerson() {

  fb3->Clear(0xFFFFFFFF, 0.0f);
  for (int tmi = 0; tmi < tmeshesN; tmi++) {
    tmeshes[tmi].RenderWireframe(ppc3, fb3);
  }
  float actualFocalLength = 30.0f;
  ppc->Draw(actualFocalLength, fb3, ppc3);
  ppc0->Draw(actualFocalLength, fb3, ppc3);
  ppc1->Draw(actualFocalLength, fb3, ppc3);
  fb3->redraw();

}

// function linked to the DBG GUI button for testing new features
void Scene::DBG() {


  {

    int u0 = 20;
    int v0 = 50;
    Kernel krn;
//    krn.SetAverage(17);
//    krn.Normalize();
    krn.SetEdgeDetection();
    FrameBuffer *cvfb = new FrameBuffer(u0+fb->w+u0, v0, fb->w, fb->h);
    cvfb->label("Convolution Output");
    krn.Convolve(fb, cvfb);
    cvfb->show();
    return;

  }

  {

    RenderRT(rtfb, ppc);
    rtfb->redraw();
    return;

  }

  {
    int stepsN = 1000;
    for (int si = 0; si < stepsN; si++) {
      animf = (float) si / (float) (stepsN-1);
      gpufb->redraw();
      Fl::check();
    }
    return;
  }

  {
    int stepsN = 10000;
    for (int si = 0; si < stepsN; si++) {
      ppc->C = ppc->C + V3(0.0f, 0.0f, 10.0f);
      Render(fb, ppc);
      Fl::check();
    }
    return;
  }

  {

    int stepsN = 100;
    for (int si = 0; si < stepsN; si++) {
      ppc->SetByInterpolation(ppc0, ppc1, (float)si / (float)(stepsN-1));
      Render();
      hwfb->redraw();
      Fl::check();
    }

    return;
  }


  {

    hwfb->redraw();
    return;

  }

  {
#if 1
    lfb->LoadTiff("mydbg/left.tif");
    lfb->redraw();
    rfb->LoadTiff("mydbg/right.tif");
    rfb->redraw();
    cerr << lfb->w << " " << lfb->h << endl;

    lppc = new PPC(65.0f, lfb->w, lfb->h);
    rppc = new PPC(*lppc);

#else
    // simulate data acquisition
    float theta = -7.0f;
    lppc->Pan(-theta);
    Render(lfb, lppc);

    rppc->Pan(theta);
    rppc->Tilt(3.0f);
    Render(rfb, rppc);

#endif

    // now forget that the actual pan angle (and cameras) are known, and try to find
    // them through optimization
    *lppc = *ppc;
    *rppc = *lppc;
    float panAngle; // current guess for pan angle
    float panAngle0 = -5.0f;
    float panAngle1 = -20.0f;
    float panStep = -1.0f;
    float tiltAngle; // current guess for tilt angle
    float tiltAngle0 = -5.0f;
    float tiltAngle1 = 5.0f;
    float tiltStep = 1.0f;
    float rollAngle; // current guess for roll angle
    float rollAngle0 = -5.0f;
    float rollAngle1 = 5.0f;
    float rollStep = 1.0f;
    float bestPanAngle, bestTiltAngle, bestRollAngle, bestError = FLT_MAX;
    for (panAngle = panAngle0; panAngle >= panAngle1; panAngle += panStep) {
      for (tiltAngle = tiltAngle0; tiltAngle <= tiltAngle1; tiltAngle += tiltStep) {
        for (rollAngle = rollAngle0; rollAngle <= rollAngle1; rollAngle += rollStep) {
          *rppc = *ppc;
          rppc->Pan(panAngle);
          rppc->Tilt(tiltAngle);
          rppc->Roll(rollAngle);
          cfb->Clear(0xFF00FF00, 0.0f);
          cfb->DrawFrameBufferSameEye(lfb, lppc, cppc);
          float error = cfb->DrawFrameBufferSameEye(rfb, rppc, cppc);
          if (error < bestError) {
            bestPanAngle = panAngle;
            bestTiltAngle = tiltAngle;
            bestRollAngle = rollAngle;
            bestError = error;
            cerr << endl << "Found smaller error: " << bestError << endl;
          }
          cerr << "Color difference: " << error << "          \r";
          cfb->redraw();
          Fl::check();
        }
      }
    }
    *rppc = *ppc;
    rppc->Pan(bestPanAngle);
    rppc->Tilt(bestTiltAngle);
    rppc->Roll(bestRollAngle);
    cfb->Clear(0xFF00FF00, 0.0f);
    cfb->DrawFrameBufferSameEye(lfb, lppc, cppc);
    cfb->DrawFrameBufferSameEye(rfb, rppc, cppc);
    cfb->redraw();
    cerr << endl;

    return;

  }

  {

    Render();
    return;
  }

  {

    int stepsN = 100;
    for (int si = 0; si < stepsN; si++) {
      ppc->SetByInterpolation(ppc0, ppc1, (float)si / (float)(stepsN-1));
      Render();
      fb3->Clear(0xFFFFFFFF, 0.0f);
      fb3->DrawFrameBuffer(fb, ppc, 1, ppc3);
      Fl::check();
    }

    Render();
    return;
  }

  {
    float theta = 1.0f;
    for (int i = 0; i < 30; i++) {
      fb->Clear(0xFFFFFFFF, 0.0f);
      Render();
      ppc->Pan(theta);
      fb->redraw();
      Fl::check();
    }
    return;

  }

  {
    V3 P(0.0f, 20.0f, -100.0f);
    V3 red(1.0f, 0.0f, 0.0f);
    V3 aDir(0.0f, 1.0f, 0.0f);
    V3 Oa(0.0f, 0.0f, -80.0f);
    V3 black(0.0f, 0.0f, 0.0f);
    int pointSize = 13;
    fb->Clear(0xFFFFFFFF, 0.0f);
    fb->Draw3DSegment(Oa, Oa + aDir*50.0f, black, red, ppc);
    for (int i = 0; i < 360; i++) {
      P = P.RotateThisPointAboutAxis(Oa, aDir, 1.0f);
      fb->Draw3DPoint(P, red, pointSize, ppc);
      fb->redraw();
      Fl::check();
    }
    return;
  }

  {
    V3 P(0.0f, 0.0f, -100.0f);
    V3 color(1.0f, 0.0f, 0.0f);
    fb->Clear(0xFFFFFFFF, 0.0f);
    int pointSize = 7;
    fb->Draw3DPoint(P, color, pointSize, ppc);
    fb->redraw();
    return;
  }


  for (int fi = 0; fi < 100; fi++) {
    Render();
    Fl::check();
    tmeshes[0].Translate(V3(0.0f, 2.0f, 0.0f));
  }
  return;

  fb->Clear(0xFFFFFFFF, 0.0f);
  V3 P0(-20.0f, 0.0f, -100.0f);
  V3 P1(+20.0f, 0.0f, -100.0f);
  V3 C0(1.0f, 0.0f, 0.0f);
  V3 C1(1.0f, 0.0f, 0.0f);
  fb->Draw3DSegment(P0, P1, C0, C1, ppc);
  fb->redraw();
  return;

  V3 v(1.0f, 2.0f, 3.0f);
  v[1] = 13.0f;
  cerr << "v: " << v << endl;
  return;

  unsigned int bgr = 0xFFFFFFFF;
  fb->Clear(bgr, 0.0f);
  fb->redraw();

//  cerr << "INFO: pressed DBG button" << endl;

  return;

}

void Scene::RenderHW() {

  // clear HW framebuffer
  glEnable(GL_DEPTH_TEST);
  glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  ppc->SetIntrinsicsHW();
  ppc->SetExtrinsicsHW();
  for (int tmi = 0; tmi < tmeshesN; tmi++) {
    tmeshes[tmi].RenderHW();
  }

}

void Scene::RenderGPU() {

  if (!cgi) {
    cgi = new CGInterface();
    soi = new ShaderOneInterface();
    cgi->PerSessionInit();
    soi->PerSessionInit(cgi);
  }
  
  cgi->EnableProfiles();
  soi->BindPrograms();
  soi->PerFrameInit();

  RenderHW();

  soi->PerFrameDisable();
  cgi->DisableProfiles();

}